(*<*)
theory VDM = IMP:
(*>*)

section{*VDM logic*}
text{*\label{sec:VDM} Our program logic is a logic for partial
correctness.*}

subsection{*Assertions, judgements, and proof rules*}
text{* As states in IMP are total functions and the language does not
include heap operations, programs can't get stuck and a VDM logic in
\emph{precondition-less style} suffices. Hence, assertions (also
called \emph{specifications}) are binary relations over states:*}

types "VDMAssn" = "State \<Rightarrow> State \<Rightarrow> bool"

text{*The program logic already includes the procedure call
instruction and thus uses proof contexts -- sets of specifications,
all interpreted with respect to the single procedure:*}

types "VDMCtxt" = "VDMAssn set"

text{*The type of the judgements, plus some pretty-printing stuff.*}

consts VDM_proof :: "(VDMAssn set \<times> IMP \<times> VDMAssn) set"
syntax VDM_deriv :: "[VDMAssn set, IMP, VDMAssn] \<Rightarrow> bool"
                     ("_ \<rhd> _ : _" [100,100,100] 100)
translations "G \<rhd> C : A" == "(G,C,A) \<in> VDM_proof"

text{*The proof rules are as follows. Apart from the final two rules,
the logic is the same as in the previous formalization. *}

inductive VDM_proof intros 
VDMSkip:   "G \<rhd> Skip : (\<lambda> s t . t=s)"
VDMAssign: "G \<rhd> (Assign x e) : (\<lambda> s t . t = (update s x (evalE e s)))"
VDMComp:   "\<lbrakk> G \<rhd> C1:A1; G \<rhd> C2:A2\<rbrakk> 
           \<Longrightarrow> G \<rhd> (Comp C1 C2) : (\<lambda> s t . \<exists> r . A1 s r \<and> A2 r t)"
VDMIff:    "\<lbrakk> G \<rhd> C1 : A; G \<rhd> C2 :B\<rbrakk>
            \<Longrightarrow> G \<rhd> (Iff b C1 C2) : (\<lambda> s t . (((evalB b s) \<longrightarrow> A s t) \<and> 
                                             ((\<not> (evalB b s)) \<longrightarrow> B s t)))"
VDMWhile_Hoare:
   "G \<rhd> C : (\<lambda> s s' . \<forall> (r::State) . evalB b s \<longrightarrow> I s r \<longrightarrow> I s' r) \<Longrightarrow>
    G \<rhd> While b C : (\<lambda> s s'. \<forall> r . I s r \<longrightarrow> (I s' r \<and> \<not> evalB b s'))"
VDMCall: "({A} \<union> G) \<rhd> body : A \<Longrightarrow> G \<rhd> Call : A"
VDMAx: "A \<in> G \<Longrightarrow>  G \<rhd> Call : A"
VDMConseq: "\<lbrakk> G \<rhd> C : A; \<forall> s t. A s t \<longrightarrow> B s t\<rbrakk> \<Longrightarrow> G \<rhd> C : B"
(*The rule is certainly valid, but we haven't derived in in
  VDMinHHoareyet, so we leave it out:
  VDMOpsem: "G \<rhd> C : (\<lambda> s t . s,C \<Down> t)"
*)
VDMTrue: "G \<rhd> C : (\<lambda> s t . True)"
VDMInter:"\<lbrakk>G \<rhd> C : A1;  G \<rhd> C : A2;  A = (\<lambda> s t. A1 s t \<and> A2 s t)\<rbrakk> 
    \<Longrightarrow> G \<rhd> C : A"  


subsection{*Admissible rules*}

text{*The while-rule is in fact inter-derivable with the following
rule.*}

lemma VDMWhile:
  "\<lbrakk> G \<rhd> C : B; \<forall> s . (\<not> evalB b s) \<longrightarrow> A s s;
     \<forall> s r t. evalB b s \<longrightarrow> B s r \<longrightarrow> A r t \<longrightarrow> A s t \<rbrakk>
  \<Longrightarrow> G \<rhd> (While b C) : (\<lambda> s t . A s t \<and> \<not> (evalB b t))"
(*<*)
apply (rule VDMConseq)
apply (rule VDMWhile_Hoare [of G C b "\<lambda> s r . \<forall> t . A s t \<longrightarrow> A r t"])
apply (erule VDMConseq) apply clarsimp
apply fast
done
(*>*)

text{*Here's the proof in the opposite direction.*}

lemma 
   "G \<rhd> C : (\<lambda> s s' . \<forall> (r::State) . evalB b s \<longrightarrow> I s r \<longrightarrow> I s' r) \<Longrightarrow>
    G \<rhd> While b C : (\<lambda> s s'. \<forall> r . I s r \<longrightarrow> (I s' r \<and> \<not> evalB b s'))"
(*<*)
apply (subgoal_tac "G \<rhd> (While b C) : (\<lambda> s t . (\<lambda> s t . \<forall>r. I s r \<longrightarrow> I t r) s t \<and> \<not> (evalB b t))")
apply (erule VDMConseq)
apply simp
apply (rule VDMWhile) apply assumption apply simp apply simp
done
(*>*)

text{*From VDMConseq and VDMInter we get the obvious rule for the
union of specifications.*}

lemma VDMUnion:
  "\<lbrakk>G \<rhd> C : A1;  G \<rhd> C : A2;  A = (\<lambda> s t. A1 s t \<or> A2 s t)\<rbrakk> 
    \<Longrightarrow> G \<rhd> C : A"
(*<*)
apply clarsimp  
apply (rule VDMConseq)
apply (erule VDMInter) apply assumption apply simp
apply fast
done
(*>*)

text{*A weakening rule and some cut rules are easily derived.*}

lemma VDM_CtxtSub[rule_format]:"G \<rhd> C : A \<Longrightarrow> \<forall> H . G \<subseteq> H \<longrightarrow> H \<rhd> C : A"
(*<*)
apply (erule VDM_proof.induct)
apply clarsimp apply (rule VDMSkip)
apply clarsimp apply (rule VDMAssign)
apply clarsimp apply (erule_tac x=H in allE, clarsimp)+ 
  apply (rule VDMComp) apply assumption apply assumption
apply clarsimp apply (erule_tac x=H in allE, clarsimp)+ 
  apply (rule VDMIff) apply assumption apply assumption
apply clarsimp apply (erule_tac x=H in allE, clarsimp)+ 
  apply (rule VDMWhile_Hoare) apply clarsimp 
apply clarsimp apply (erule_tac x="{A} \<union> H" in allE, erule impE) apply fastsimp
  apply (erule VDMCall) 
apply clarsimp apply (rule VDMAx) apply fast
apply clarsimp apply (erule_tac x=H in allE, clarsimp) 
  apply (erule VDMConseq) apply assumption
(*apply clarsimp apply (rule VDMOpsem)*)
apply clarsimp apply (rule VDMTrue)
apply clarsimp apply (erule_tac x=H in allE, erule impE, assumption)+ 
  apply (erule VDMInter)  apply assumption apply simp
done
(*>*)

(*<*)
constdefs CutAuxProp::"VDMAssn set \<Rightarrow> IMP \<Rightarrow> VDMAssn \<Rightarrow> bool"
"CutAuxProp DD C A \<equiv> 
  \<forall> G P D . (DD = (insert P D) \<longrightarrow> G \<rhd> Call :P \<longrightarrow> (G \<subseteq> D) \<longrightarrow> D \<rhd> C:A)"

lemma CutAux1:"(DD \<rhd> C : A) \<Longrightarrow> CutAuxProp DD C A"
apply (erule VDM_proof.induct)
apply (simp_all add: add: CutAuxProp_def)
apply clarify
apply (fast intro: VDMSkip)
apply (fast intro: VDMAssign)
apply clarsimp 
  apply (erule_tac x=Ga in allE) apply (erule_tac x=Ga in allE)
  apply (erule_tac x=P in allE) apply (erule_tac x=P in allE)
  apply (erule_tac x=D in allE, simp) apply (erule_tac x=D in allE, simp)
  apply (rule VDMComp) apply assumption+
apply clarsimp 
  apply (erule_tac x=Ga in allE) apply (erule_tac x=Ga in allE)
  apply (erule_tac x=P in allE) apply (erule_tac x=P in allE)
  apply (erule_tac x=D in allE, simp) apply (erule_tac x=D in allE, simp)
  apply (rule VDMIff) apply assumption+
apply clarsimp 
  apply (erule_tac x=Ga in allE) 
  apply (erule_tac x=P in allE) 
  apply (erule_tac x=D in allE, simp) 
  apply (rule VDMWhile_Hoare) apply clarsimp
apply clarsimp 
  apply (erule_tac x=Ga in allE) 
  apply (erule_tac x=P in allE) 
  apply (erule_tac x="insert A D" in allE, simp) 
  apply (rule VDMCall) apply fastsimp
apply clarsimp
  apply (erule disjE)
  apply clarsimp apply (erule VDM_CtxtSub) apply assumption
  apply (erule VDMAx)
apply clarsimp
apply (subgoal_tac "D \<rhd>  C : A") 
apply (erule VDMConseq) apply assumption  
  apply simp
apply clarsimp
  apply (rule VDMTrue)
apply clarsimp 
  apply (erule_tac x=Ga in allE, erule_tac x=P in allE, erule_tac x=D in allE, erule impE, simp) 
  apply (erule_tac x=Ga in allE, erule_tac x=P in allE, erule_tac x=D in allE, erule impE, simp) 
  apply simp 
  apply (erule VDMInter) apply assumption apply rule apply rule apply fast
done
(*>*)

lemma CutAux: 
  "\<lbrakk>DD \<rhd> C : A; DD = insert P D; 
    G \<rhd> Call :P; G \<subseteq> D\<rbrakk> \<Longrightarrow> D \<rhd> C:A"
(*<*)
by (drule CutAux1, simp add: CutAuxProp_def)
(*>*)

lemma cut:"\<lbrakk> G \<rhd> Call : P ; (insert P G) \<rhd> C : A \<rbrakk> \<Longrightarrow> G \<rhd> C : A"
(*<*)
apply (rotate_tac -1, drule CutAux)
apply (fastsimp, assumption)
apply (simp, assumption)
done
(*>*)

text{*We call context $G$ verified if all entries are justified by
derivations for the procedure body.*}

constdefs verified::"VDMAssn set \<Rightarrow> bool"
"verified G \<equiv> (\<forall> A . A:G \<longrightarrow> G \<rhd> body : A)"

text{*The property is preserved by sub-contexts*}

lemma verified_preserved: "\<lbrakk>verified G; A:G\<rbrakk> \<Longrightarrow> verified (G - {A})"
(*<*)
apply (simp add: verified_def, (rule allI)+, rule)
apply (subgoal_tac "(G - {A}) \<rhd> Call:A")
(*now use the subgoal (G - {A}) \<rhd> Call:A to prove the goal*)
apply (subgoal_tac "G = insert A (G - {A})") prefer 2 apply fast
apply (erule_tac x=Aa in allE) 
apply (erule impE, simp)
apply (erule cut)  apply simp
(* now prove the subgoal (G - {A}) \<rhd>  Call : A *)
  apply (erule_tac x=A in allE, clarsimp)
  apply (rule VDMCall) apply simp apply (erule VDM_CtxtSub) apply fast
done
(*>*)

(*<*)
lemma SetNonSingleton:
  "\<lbrakk>G \<noteq> {A}; A \<in> G\<rbrakk> \<Longrightarrow> \<exists>B. B \<noteq> A \<and> B \<in> G"
by auto

lemma MUTREClemma[rule_format]: 
"\<forall> G . ((finite G \<and> card G = n \<and> verified G \<and> A : G) \<longrightarrow> {} \<rhd> Call:A)"
apply (induct n)
(*case n=0*)
apply clarsimp
(*Case n>0*)
apply clarsimp
apply (case_tac "G = {A}")
apply clarsimp
apply (erule_tac x="{A}" in allE)
apply (clarsimp, simp add:verified_def)
apply (rule VDMCall, clarsimp)
(*Case G has more entries than A*)
apply (drule SetNonSingleton, assumption) 
(* use the fact that there is another entry B:G*)
apply clarsimp
apply (subgoal_tac "(G - {B}) \<rhd> Call : B")
apply (erule_tac x="G - {B}" in allE)
apply (erule impE) apply (simp add: verified_preserved)
apply (erule impE) apply (simp add: card_Diff_singleton)
apply simp
(*the proof for (G - {B}) \<rhd>  Call : B *)
apply (simp add: verified_def)
apply (rotate_tac 3) apply (erule_tac x=B in allE, simp)
apply (rule VDMCall) apply simp apply (erule VDM_CtxtSub) apply fast
done
(*>*)

text{*The \emph{Mutrec} rule allows to eliminate verified finite
contexts.*}

theorem MUTREC:
"\<lbrakk> finite G; card G = n; verified G; A : G \<rbrakk> \<Longrightarrow> {} \<rhd> Call:A"
(*<*) 
by (rule MUTREClemma, fast)
(*>*)

subsection{*Interpretation*}
text{*The validity of an assertion with respect to a program is
defined as the naive partial-correctness statement: *}

constdefs vdm_valid :: "IMP \<Rightarrow> VDMAssn \<Rightarrow> bool"
                       (" \<Turnstile> _ : _ " [100,100] 100)
"vdm_valid C A \<equiv> \<forall> s t . (s,C \<Down> t) \<longrightarrow> A s t"

text{*A context is valid if all its entries are valid for the
procedure call instruction. Obviously, empty contexts are trivially
valid.*}

constdefs vdmctxt_valid :: "VDMCtxt \<Rightarrow> bool"
                        (" \<Turnstile> _ " [100] 100)
"vdmctxt_valid G \<equiv> \<forall> A . A \<in> G \<longrightarrow> vdm_valid Call A"  

text{*Finally, a judgement is valid if the validity of the context
implies that of the specification.*}

constdefs v_valid::  "VDMCtxt \<Rightarrow> IMP \<Rightarrow> VDMAssn \<Rightarrow> bool"
                    ("_ \<Turnstile> _ : _ " [100,100,100] 100)
"v_valid G C A \<equiv> vdmctxt_valid G \<longrightarrow> vdm_valid C A"

(*<*)
end
(*>*)